/*
 * Based on BluetoothChat sample
 * Original code: https://android.googlesource.com/platform/development/+/master/samples/BluetoothChat/src/com/example/android/BluetoothChat/BluetoothChatService.java?autodive=0%2F
 * Original author copyright declared below
 */

/*
 * Copyright (C) 2009 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package com.megabras.bluelogg.bluetooth;

import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.Context;
import android.os.Handler;

import com.megabras.modbus.MUtils;
import com.megabras.modbus.Message;
import com.megabras.modbus.Modbus;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.LinkedList;
import java.util.Objects;
import java.util.UUID;

public class BluetoothClassic extends BluetoothDev
{

    private static final UUID
            MY_UUID = UUID.fromString( "00001101-0000-1000-8000-00805f9b34fb" );

    private ConnectThread
            mConnectThread;

    private ConnectedThread
            mConnectedThread;

    private boolean
            mustStop;

    public BluetoothClassic( Context context, BluetoothDevice device, Handler handler, Message messageStatus )
    {
        super( context, device, handler, messageStatus);
        mustStop = false;
    }

    /**
     * Start the ConnectThread to initiate a connection to a remote device.
     *
     * @param secure Socket Security type - Secure (true) , Insecure (false)
     */
    @Override
    public synchronized void connect( boolean secure )
    {
        super.connect( secure );
        // Cancel any thread attempting to make a connection
        if ( getState() == STATE_CONNECTING )
        {
            if ( mConnectThread != null )
            {
                mConnectThread.cancel();
                mConnectThread = null;
            }
        }

        // Cancel any thread currently running a connection
        if ( mConnectedThread != null )
        {
            mConnectedThread.cancel();
            mConnectedThread = null;
        }

        // Start the thread to connect with the given device
        mConnectThread = new ConnectThread( device, secure );
        mConnectThread.start();
    }

    /**
     * Start the ConnectedThread to begin managing a Bluetooth connection
     *
     * @param socket The BluetoothSocket on which the connection was made
     */
    public synchronized void connected( BluetoothSocket socket )
    {
        // Cancel the thread that completed the connection
        if ( mConnectThread != null )
        {
            mConnectThread.cancel();
            mConnectThread = null;
        }

        // Cancel any thread currently running a connection
        if ( mConnectedThread != null )
        {
            mConnectedThread.cancel();
            mConnectedThread = null;
        }

        // Start the thread to manage the connection and perform transmissions
        mConnectedThread = new ConnectedThread( socket );
        mConnectedThread.start();

        setState( STATE_CONNECTED );
    }

    /**
     * Stop all threads
     */
    @Override
    public synchronized void stop()
    {
        super.stop();
        mustStop = true;
        if ( mConnectThread != null )
        {
            mConnectThread.cancel();
            mConnectThread = null;
        }

        if ( mConnectedThread != null )
        {
            mConnectedThread.cancel();
            mConnectedThread = null;
        }
    }

    /**
     * Write to the ConnectedThread in an unsynchronized manner
     *
     * @param out The bytes to write
     * @see ConnectedThread#write(byte[])
     */
    public void write(byte[] out) {
        // Create temporary object
        ConnectedThread r;
        // Synchronize a copy of the ConnectedThread
        synchronized (this) {
            if (getState() != STATE_CONNECTED && getState() != STATE_NOT_RESPONDING) return;
            r = mConnectedThread;
        }
        // Perform the write unsynchronized
        r.write(out);
    }

    @Override
    public void addMessage( Message message )
    {
        // Create temporary object
        ConnectedThread r;
        // Synchronize a copy of the ConnectedThread
        synchronized ( this )
        {
            if ( getState() != STATE_CONNECTED && getState() != STATE_NOT_RESPONDING )
                return;
            r = mConnectedThread;
        }
        // Perform the write unsynchronized
        r.addMessage( message );
    }

    @Override
    public void setTimeout(int timeout){
        // Create temporary object
        ConnectedThread r;
        // Synchronize a copy of the ConnectedThread
        synchronized (this) {
            r = mConnectedThread;
        }
        // Perform the write unsynchronized
        r.setTimeout(timeout);
    }

    @Override
    public void setCycleTime(int time)
    {
        // Create temporary object
        ConnectedThread r;
        // Synchronize a copy of the ConnectedThread
        synchronized (this) {
            r = mConnectedThread;
        }
        // Perform the write unsynchronized
        r.setCycleTime(time);
    }

    @Override
    public void priorityMessage(Message message) {
        // Create temporary object
        ConnectedThread r;
        // Synchronize a copy of the ConnectedThread
        synchronized (this) {
            if (getState() != STATE_CONNECTED && getState() != STATE_NOT_RESPONDING) return;
            r = mConnectedThread;
        }
        // Perform the write unsynchronized
        r.priorityMessage(message);
    }

    /**
     * Indicate that the connection attempt failed and notify the UI Activity.
     */
    private void connectionFailed(String error) {
        setState(STATE_FAIL, error);
    }

    /**
     * Indicate that the connection was lost and notify the UI Activity.
     */
    @SuppressWarnings("SameParameterValue")
    private void connectionLost( String error )
    {
        setState( STATE_LOST, error );
    }

    private void connectionNotResponding()
    {
        setState(STATE_NOT_RESPONDING);
    }

    /**
     * This thread runs while attempting to make an outgoing connection
     * with a device. It runs straight through; the connection either
     * succeeds or fails.
     */
    private class ConnectThread extends Thread {
        private BluetoothSocket
                mmSocket;

        private final BluetoothDevice
                mmDevice;

        private final boolean
                secure;

        ConnectThread(BluetoothDevice device, boolean secure) {
            mmDevice = device;
            this.secure = secure;
        }

        public void run() {
            try {
                if(secure){
                    mmSocket = mmDevice.createRfcommSocketToServiceRecord(MY_UUID);
                }
                else {
                    mmSocket = mmDevice.createInsecureRfcommSocketToServiceRecord(MY_UUID);
                }
            }
            catch (Exception e) {
                connectionFailed("ConnectThread-1");
                return;
            }

            // Make a connection to the BluetoothSocket
            try {
                // This is a blocking call and will only return on a successful connection or an exception
                mmSocket.connect();
            }
            catch ( Exception e )
            {
                // Close the socket
                try
                {
                    if ( mmSocket != null )
                        mmSocket.close();
                }
                catch ( IOException ignored )
                {
                }
                connectionFailed( "ConnectThread-2" );
                return;
            }

            // Reset the ConnectThread because we're done
            synchronized ( BluetoothClassic.this) {
                mConnectThread = null;
            }

            // Start the connected thread
            connected(mmSocket);
        }

        public void cancel() {
            try {
                mmSocket.close();
                mmSocket = null;
            }
            catch (IOException ignore) {}
        }
    }

    /**
     * This thread runs during a connection with a remote device.
     * It handles all incoming and outgoing transmissions.
     */
    private class ConnectedThread extends Thread {

        private final BluetoothSocket
                mmSocket;

        private InputStream
                mmInStream;

        private OutputStream
                mmOutStream;

        ConnectedThread(BluetoothSocket socket) {
            mmSocket            = socket;
            InputStream tmpIn   = null;
            OutputStream tmpOut = null;

            // Get the BluetoothSocket input and output streams
            try {
                tmpIn  = mmSocket.getInputStream();
                tmpOut = mmSocket.getOutputStream();
            }
            catch (IOException ignore) {}

            mmInStream   = tmpIn;
            mmOutStream  = tmpOut;
            FIFO_MESSAGE = new LinkedList<>();
            runFIFO      = new Runnable() {
                public void run() {
                    clycleMessage();
                    mHandler.postDelayed(this, CYCLE_TIME);
                }
            };

            mHandler.removeCallbacks(runFIFO);
            mHandler.postDelayed(runFIFO, CYCLE_TIME);
        }

        private void clycleMessage()
        {
            if (idle) // COMUNICATION: IDLE
            {
                if (!FIFO_MESSAGE.isEmpty())
                {
                    idle         = false;
                    rx           = "";
                    write( (Objects.requireNonNull( FIFO_MESSAGE.peek() ).getMessage()).getBytes() );
                    current_time = 0;
                }
            }
            else // COMUNICATION: WAITING FOR RESPONSE
            {
                if (checkResponseWithFrames(rx) || checkResponseWithoutFrames(rx))
                {
                    rx = rx.replace(Modbus.START_FRAME,"").replace(Modbus.START_FRAME,"");

                    if (FIFO_MESSAGE.peek()==null)
                        return;

                    String tx = Objects.requireNonNull( FIFO_MESSAGE.peek() ).getMessage().replace(Modbus.START_FRAME,"").replace(Modbus.START_FRAME,"");

                    boolean isExpectedAnswer = false;
                    String rxFunctionCode    = rx.substring(2,4);
                    String txFunctionCode    = tx.substring(2,4);

                    if (rxFunctionCode.equals(txFunctionCode))
                    {
                        if (rxFunctionCode.equals("03"))
                        {
                            int txSize = MUtils.hexToInt(tx.substring(8,12));
                            int rxSize = MUtils.hexToInt(rx.substring(4,6))/2;
                            if (txSize==rxSize)
                                isExpectedAnswer = true;
                        }
                        else
                        {
                            String txAddress = tx.substring(4,8);
                            String rxAddress = rx.substring(4,8);
                            if (rxAddress.equals(txAddress))
                                isExpectedAnswer = true;
                        }
                    }

                    if (isExpectedAnswer)
                    {
                        if (FIFO_MESSAGE.peek()!=null)
                        {
                            mHandler.obtainMessage( MESSAGE_READ, 0, Objects.requireNonNull( FIFO_MESSAGE.peek() ).getID(), rx).sendToTarget();
                            FIFO_MESSAGE.remove();
                            current_time = 0;
                            idle = true;
                            rx = "";
                        }
                    }
                }
            }

            // COMUNICATION: Rx TIMEOUT and CHECK CONNECTION
            if (current_time==TIME_OUT) // Retry last command OR ask status
            {
                if (FIFO_MESSAGE.isEmpty())
                    FIFO_MESSAGE.add( messageStatus );
                rx           = "";
                idle         = false;
                if (FIFO_MESSAGE.peek()!=null)
                    write( (Objects.requireNonNull( FIFO_MESSAGE.peek() ).getMessage()).getBytes() );
            }
            else if (current_time==(TIME_OUT*2))
            {
                if (FIFO_MESSAGE.isEmpty())
                    FIFO_MESSAGE.add( messageStatus );
                rx           = "";
                idle         = false;
                if (FIFO_MESSAGE.peek()!=null)
                    write( (Objects.requireNonNull( FIFO_MESSAGE.peek() ).getMessage()).getBytes() );
            }
            else if (current_time==(TIME_OUT*4))
            {
                connectionNotResponding();
                idle         = true;
                rx           = "";
                current_time = 0;
            }

            current_time += CYCLE_TIME;
        }

        void addMessage(Message message){
            FIFO_MESSAGE.add(message);
        }

        public void setTimeout(int timeout){
            TIME_OUT = timeout;
        }

        void setCycleTime(int time)
        {
            CYCLE_TIME = time;
        }

        void priorityMessage(Message message)
        {
            idle         = true;
            current_time = 0;
            FIFO_MESSAGE.clear();
            FIFO_MESSAGE.add( message );
        }

        public void run()
        {
            byte[] buffer = new byte[1024];
            int bytes;

            while ( !mustStop )
            {
                try
                {
                    bytes               = mmInStream.read( buffer );
                    String readMessage  = new String( buffer, 0, bytes );
                    rx                 += readMessage;
                    current_time        = 0;
                }
                catch ( IOException e )
                {
                    mHandler.removeCallbacks( runFIFO );
                    if ( !mustStop )
                    {
                        connectionLost( "ConnectedThread" );
                    }
                    break;
                }
            }
        }

        /**
         * Write to the connected OutStream.
         *
         * @param buffer The bytes to write
         */
        public void write( byte[] buffer )
        {
            try
            {
                if ( mmOutStream != null && FIFO_MESSAGE.peek() != null )
                {
                    int id = Objects.requireNonNull( FIFO_MESSAGE.peek() ).getID();
                    mmOutStream.write( buffer );
                    mHandler.obtainMessage( MESSAGE_WRITE, -1, id, new String( buffer ).replace( ":", "" ).replace( "\r\n", "" ) ).sendToTarget();
                }
            }
            catch ( IOException ignore )
            {
            }
        }

        public void cancel()
        {
            try
            {
                mHandler.removeCallbacks( runFIFO );
                if ( mmInStream != null )
                {
                    mmInStream.close();
                    mmInStream = null;
                }
                if ( mmOutStream != null )
                {
                    mmOutStream.close();
                    mmOutStream = null;
                }

                mmSocket.close();
            }
            catch ( Exception ignore )
            {
            }
        }
    }
}

